#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "m3_SerialClient.h"
#include "m3_defines.h"

#define DSP_ENTER_MONITOR 0x95
#define DSP_LEAVE_MONITOR 0xA5
#define DSP_GET_STATUS 0x53
#define DSP_GET_CONFIG 0x33
#define DSP_DO_FLASH 0xC5
#define DSP_ERASE_FLASH 0x33
#define DSP_ERASE_STATUS 0x53
#define DSP_PGM_FLASH 0x93
#define DSP_READ_FLASH 0xA3

int read_s19(int *buf, char *filename) {
  char line[512];
  FILE *f;
  int len;
  int addr;
  int i, done = 0;
  int mysum, checksum;
  char *c;
  
  f = fopen(filename,"r");
  if (f== NULL) {
    printf("Error. can't open %s\n",filename);
    return(-1);
  }
  while (!done) {
    fgets(line,512,f);
    if (!strncmp(line,"S3",2)) {
      c = line+2;
      len =-1;
      sscanf(c,"%2x",&len);
      if (len == -1) {
	printf("Error with length field\n");
	return(-1);
      }
      mysum = len;
      c +=2;
      sscanf(c,"%8x",&addr);
      mysum += (addr>>24)&0xFF;
      mysum += (addr>>16)&0xFF;
      mysum += (addr>>8)&0xFF;
      mysum += addr&0xFF;
      c+=8;
      for(i=0; i<len-5; i++) {
	if (addr+i > 0x7FFFFF) { printf("OutOfRange\n");
	exit(-1);
	}
	sscanf(c,"%2x",buf+addr+i);
	c +=2;
	mysum += buf[addr+i];
      }
      sscanf(c,"%2x",&checksum);
      if (checksum != ((mysum&0xFF)^0xFF)) {
	printf("Checksum error on line:\n%s\n",line);
	return(-1);
      }
    } else if (!strncmp(line,"S9",2)) {
      done = 1;
    } else if (!strncmp(line,"S7",2)) {
      done = 1;
    }
    if (feof(f)) {
      printf("Missing S7/S9 termination record\n");
      return(-1);
    }
  }
  return(0);
}

int get_num_linecards(m3_SerialClient *client) {
  m3_serial_command_t cmd;
  char txbuf[8];
  unsigned char rxbuf[1024];
  int tries;
  int i;
  char *c;
  int subtype;

  tries = 0;
  while (tries<3) {
    cmd.command = DSP_GET_STATUS;
    cmd.port = 55;
    cmd.txlen = 0;
    cmd.rxlen = 1000;
    cmd.txbuf = txbuf;
    cmd.rxbuf = (char*)(rxbuf);
    cmd.result = 0;
    
    if ((client->run(1,&cmd)) &&
	(cmd.result >= 16)) break;
#if 0
    printf("Couldn't talk to linecard, result= %d. Trying again.\n",
	   cmd.result);
#endif
    usleep(500000);   
    tries++;
  }
  if (tries >=3) {
    printf("Error - can't talk to linecard. Check IP address/hostname\n");

    return -1;
  }
  if (rxbuf[9] ==2) return 2;
  if (rxbuf[9] ==4) return 4;
  if (rxbuf[9] ==8) return 8;
  if (rxbuf[9] ==16) return 16;
  printf("Unknown backplane: %02x\n",rxbuf[9]);
  exit(-1);
}


#define M3_ERROR  -1
#define M3_UNKNOWN 0
#define M3_SW16_8E 1
#define M3_SW16_8P 2

int get_linecard_type(m3_SerialClient *client, int port) {
  m3_serial_command_t cmd;
  char txbuf[8];
  unsigned char rxbuf[1024];
  int tries;
  int i;
  char *c;
  int subtype;

  tries = 0;
  while (tries<3) {
    cmd.command = DSP_GET_CONFIG;
    cmd.port = port;
    cmd.txlen = 0;
    cmd.rxlen = 1000;
    cmd.txbuf = txbuf;
    cmd.rxbuf = (char*)(rxbuf);
    cmd.result = 0;
    
    if ((client->run(1,&cmd)) &&
	(cmd.result >= 20)) break;
#if 0
    printf("Couldn't talk to linecard, result= %d. Trying again.\n",
	   cmd.result);
#endif
    usleep(500000);   
    tries++;
  }
  if (tries >=3) {
#if 0
    printf("Error - can't talk to linecard. Check IP address/hostname and slot number\n");
#endif
    return M3_ERROR;
  }
  rxbuf[1000] = 0;
  i= rxbuf[0]+1; /* skip type bytes */
  subtype = M3_UNKNOWN;
  if (strstr((char *)rxbuf+i,"M3-SW16-8E")) subtype = M3_SW16_8E;
  if (strstr((char *)rxbuf+i,"M3-SW16-8P")) subtype = M3_SW16_8P;
  while ((rxbuf[i++]) && (i < cmd.result)); /* skip serial number */
  while ((rxbuf[i++]) && (i < cmd.result)); /* skip EDC */  
  while ((rxbuf[i++]) && (i < cmd.result)); /* skip version */
  if (i< cmd.result) {
    if (strncmp((char *)rxbuf+i,"Monitor for dsp56800e", 21)) {
      subtype = M3_UNKNOWN;
    }
  }
  return subtype;
}

int check_linecard_type(m3_SerialClient*client, int port) {
  int type;

  type = get_linecard_type(client,port);
  if (type == M3_UNKNOWN) {
    printf("Error - linecard not correct type. Check IP address/hostname and slot number\n");
    return -1;
  } else if (type == M3_ERROR) {
    printf("Error - can't talk to linecard. Check IP address/hostname and slot number\n");
    return -1;
  } else {
    return 0;
  }
}

int enter_monitor(m3_SerialClient*client, int port) {
  m3_serial_command_t cmd;
  char txbuf[8];
  unsigned char rxbuf[8];
  int tries;

  cmd.command = DSP_ENTER_MONITOR;
  cmd.port = port;
  cmd.txlen = 0;
  cmd.rxlen = 4;
  cmd.txbuf = txbuf;
  cmd.rxbuf = (char*)(rxbuf);
  cmd.result = 0;
  
  if (!(client->run(1,&cmd))) {
    fprintf(stderr,"Error talking to board1\n");
    return(-1);
  }
 
  tries = 0;
  while (tries<5) {
    rxbuf[0] = 0;
    cmd.command = DSP_GET_STATUS;
    cmd.port = port;
    cmd.txlen = 0;
    cmd.rxlen = 1;
    cmd.txbuf = txbuf;
    cmd.rxbuf = (char *)rxbuf;
    cmd.result = 0;
    if (!(client->run(1,&cmd))) {
      fprintf(stderr,"Error talking to board2\n");
      return -1;
    }
    if ((cmd.result>0)&&(rxbuf[0]==0xff)) return 0;
    usleep(500000);
    tries++;
  }
  fprintf(stderr,"Board can't enter monitor mode\n");
  return -1;
}

int leave_monitor(m3_SerialClient*client, int port) {
  m3_serial_command_t cmd;
  char txbuf[8];
  unsigned char rxbuf[8];
  int tries;

  cmd.command = DSP_LEAVE_MONITOR;
  cmd.port = port;
  cmd.txlen = 0;
  cmd.rxlen = 4;
  cmd.txbuf = txbuf;
  cmd.rxbuf = (char*)(rxbuf);
  cmd.result = 0;
  
  if (!(client->run(1,&cmd))) {
    fprintf(stderr,"Error talking to board1\n");
    return(-1);
  }
 
  tries = 0;
  while (tries<5) {
    rxbuf[0] = 0;
    cmd.command = DSP_GET_STATUS;
    cmd.port = port;
    cmd.txlen = 0;
    cmd.rxlen = 1;
    cmd.txbuf = txbuf;
    cmd.rxbuf = (char *)rxbuf;
    cmd.result = 0;
    if (!(client->run(1,&cmd))) {
      fprintf(stderr,"Error talking to board2\n");
      return -1;
    }
    if ((cmd.result>0)&&(rxbuf[0]!=0xff)) return 0;
    sleep(1);
    tries++;
  }
  fprintf(stderr,"Board can't leave monitor mode\n");
  return -1;
}

int erase_buf(m3_SerialClient*client, int port, int *buf) {
  m3_serial_command_t cmd;
  char txbuf[8];
  char rxbuf[8];
  int tries;
  int a,i,do_erase;

  for(a=0; a<0x800000; a+=0x10000) {
    do_erase = 0;
    for (i=0; i<0x10000; i++) {
      if (buf[a+i] != -1) {
	do_erase =1;
	break;
      }
    }
    if (do_erase) {
      printf("Erasing address %x\r",a);
      fflush(stdout);
      cmd.command = DSP_DO_FLASH;
      cmd.port = port;
      cmd.txlen = 6;
      cmd.rxlen = 0;
      cmd.txbuf = txbuf;
      cmd.rxbuf = rxbuf;
      txbuf[0] = DSP_ERASE_FLASH;
      txbuf[1] = DSP_ERASE_FLASH;
      txbuf[2] = a&0xff;
      txbuf[3] = (a>>8)&0xff;
      txbuf[4] = (a>>16)&0xff;
      txbuf[5] = 0;
      cmd.result = 0;
      if ( (!client->run(1,&cmd)) || (cmd.result != 0)) {
	fprintf(stderr,"Erase failed\n");
	return -1;
      }
      while (1) {
	cmd.command = DSP_DO_FLASH;
	cmd.port = port;
	cmd.txlen = 6;
	cmd.rxlen = 1;
	cmd.txbuf = txbuf;
	cmd.rxbuf = rxbuf;
	txbuf[0] = DSP_ERASE_STATUS;
	txbuf[1] = DSP_ERASE_STATUS;
	txbuf[2] = a&0xff;
	txbuf[3] = (a>>8)&0xff;
	txbuf[4] = (a>>16)&0xff;
	txbuf[5] = 0;
	cmd.result = 0;
	if ( (!client->run(1,&cmd)) || (cmd.result != 1)) {
	  fprintf(stderr,"Erase failed\n");
	  return -1;
	}
	if (rxbuf[0] == 'P') break;
	if (rxbuf[0] == 'F') {
	  fprintf(stderr,"Erase failed - flash error\n");
	  return -1;
	}
      }
    }
  }
  printf("Erase complete!            \n");
  return 0;
}

int pgm_buf(m3_SerialClient*client, int port, int *buf) {
  m3_serial_command_t *commands;
  char *txbuf;
  char *rxbuf;
  int a,i,do_prog;
  int cmd;
  commands = (m3_serial_command_t*) malloc(sizeof(m3_serial_command_t)*
					   (0x800000/128));
  
  txbuf = (char *) malloc(0x800000/128*144);
  rxbuf = (char *) malloc(0x800000/128);
  cmd = 0;
  for(a=0; a<0x800000; a+=128) {
    do_prog = 0;
    for (i=0; i<128; i++) {
      if (buf[a+i] != -1) {
	do_prog =1;
	break;
      }
    }
    if (do_prog) {
      printf("Programming address %x\r",a);
      fflush(stdout);
      commands[cmd].command = DSP_DO_FLASH;
      commands[cmd].port = port;
      commands[cmd].txlen = 135;
      commands[cmd].rxlen = 1;
      commands[cmd].txbuf = txbuf+cmd*144;
      commands[cmd].rxbuf = rxbuf+cmd;
      commands[cmd].result = 0;
      commands[cmd].txbuf[0] = DSP_PGM_FLASH;
      commands[cmd].txbuf[1] = DSP_PGM_FLASH;
      commands[cmd].txbuf[2] = a&0xff;
      commands[cmd].txbuf[3] = (a>>8)&0xff;
      commands[cmd].txbuf[4] = (a>>16)&0xff;
      commands[cmd].txbuf[5] = 0;
      commands[cmd].txbuf[6] = 128;
      for (i=0; i<128; i++) txbuf[cmd*144+i+7] = buf[a+i];
      cmd++;
    }
  }
  if (!client->run(cmd,commands)) {
    fprintf(stderr,"Program failed1\n");
    return -1;
  }
  for (i=0; i<cmd; i++) {
    if ((commands[i].result !=1)||(rxbuf[i]!='P')) {
      fprintf(stderr,"Program failed2 - %i %d\n",commands[i].result,rxbuf[i]);
      return -1;
    }
  }
  printf("Program complete!          \n");
  return 0;
}

int vfy_buf(m3_SerialClient*client, int port, int *buf) {
  m3_serial_command_t *commands;
  char *txbuf;
  unsigned char *rxbuf;
  int a,i,do_vfy;
  int cmd;
  commands = (m3_serial_command_t*) malloc(sizeof(m3_serial_command_t)*
					   (0x800000/128));
  
  txbuf = (char *) malloc(0x800000/128*8);
  rxbuf = (unsigned char *) malloc(0x800000);
  cmd = 0;
  for(a=0; a<0x800000; a+=128) {
    do_vfy = 0;
    for (i=0; i<128; i++) {
      if (buf[a+i] != -1) {
	do_vfy =1;
	break;
      }
    }
    if (do_vfy) {
      printf("Verifying address %x\r",a);
      fflush(stdout);
      commands[cmd].command = DSP_DO_FLASH;
      commands[cmd].port = port;
      commands[cmd].txlen = 7;
      commands[cmd].rxlen = 128;
      commands[cmd].txbuf = txbuf+cmd*8;
      commands[cmd].rxbuf = (char*)rxbuf+cmd*128;
      commands[cmd].result = 0;
      commands[cmd].txbuf[0] = DSP_READ_FLASH;
      commands[cmd].txbuf[1] = DSP_READ_FLASH;
      commands[cmd].txbuf[2] = a&0xff;
      commands[cmd].txbuf[3] = (a>>8)&0xff;
      commands[cmd].txbuf[4] = (a>>16)&0xff;
      commands[cmd].txbuf[5] = 0;
      commands[cmd].txbuf[6] = 128;
      cmd++;
    }
  }
  if (!client->run(cmd,commands)) {
    fprintf(stderr,"Verify failed1\n");
    return -1;
  }
  for (i=0; i<cmd; i++) {
    if (commands[i].result !=128) {
      fprintf(stderr,"Verify failed2\n");
      return -1;
    }
  }
  cmd = 0;
  for(a=0; a<0x800000; a+=128) {
    do_vfy = 0;
    for (i=0; i<128; i++) {
      if (buf[a+i] != -1) {
	do_vfy =1;
	break;
      }
    }
    if (do_vfy) {
      for (i=0; i<128; i++) {
	if ((buf[a+i]&0xFF) != (rxbuf[cmd*128+i]&0xff)) {
	  fprintf(stderr,"Verify error at address 0x%x\n",a+i);
	  printf("buf: %x != flash: %x\n",buf[a+i],rxbuf[cmd*128+i]);
	  return -1;
	}
      }
      cmd++;
    }
  }
  printf("Verify complete!                \n");
  return 0;
}


static void printLine (char*s)
{
  printf ("%s\n", s);
}

void printusage(void) {
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"upgrade_xm IP_address_or_name slot_number s19_file\n");
  fprintf(stderr,"or\n");
  fprintf(stderr,"upgrade_xm IP_address_or_name E|P s19_file\n");
}

int main (int argc, char**argv)
{
  
  if (argc != 4) {
    printusage();
    exit(-1);
  }

  m3_Component::initialize (printLine);
  m3_SerialClient client (argv[1]);

  if ((argv[2][0]=='P') || (argv[2][0]=='p') || 
      (argv[2][0]=='E') || (argv[2][0]=='e')) {
    int *buf;
    int port;
    int type;
    int lctype;
    int i;
    int num_slots;

    num_slots = get_num_linecards(&client);

    if ((argv[2][0]=='E') || (argv[2][0]=='e')) type=M3_SW16_8E;
    if ((argv[2][0]=='P') || (argv[2][0]=='p')) type=M3_SW16_8P;
    
    buf = (int *) malloc(sizeof(int)*0x800000);
    for (i=0; i<0x800000; i++) {
      buf[i] = -1;
    }

    if (read_s19(buf,argv[3])) {
      fprintf(stderr,"error reading %s\n",argv[2]);
      exit(-1);
    }
    
    printf("Enclosure has %d slots\n",num_slots+1);
    
    for (port=0; port<num_slots; port++) {
      printf("Trying slot %d...",port+1);
      fflush(stdout);
      lctype = get_linecard_type(&client,port);
      if (lctype==M3_ERROR) {
	printf(" couldn't talk to linecard.\n");
      } else  if (lctype!=type) {
	printf(" not correct type.\n");
      } else {
	printf(" OK. Programming...\n ");
	if (enter_monitor(&client,port)) { printf("Error!\n"); continue; }
	if (erase_buf(&client,port,buf)) { printf("Error!\n"); continue; }
	if (pgm_buf(&client,port,buf)) { printf("Error!\n"); continue; }
	if (vfy_buf(&client,port,buf)) { printf("Error!\n"); continue; }
	if (leave_monitor(&client,port)) { printf("Error!\n"); continue; }
	printf("OK!\n");
      }
    }
  } else if ((argv[2][0]>='0') && (argv[2][0]<='9')) {
    int port = atoi (argv[2])-1;
    int *buf;
    int i;
    
    buf = (int *) malloc(sizeof(int)*0x800000);
    for (i=0; i<0x800000; i++) {
      buf[i] = -1;
    }
    
    if (read_s19(buf,argv[3])) {
      fprintf(stderr,"error reading %s\n",argv[2]);
      exit(-1);
    }
    if (check_linecard_type(&client,port)) exit(-1);
    if (enter_monitor(&client,port)) exit(-1);
    if (erase_buf(&client,port,buf)) exit(-1);
    if (pgm_buf(&client,port,buf)) exit(-1);
    if (vfy_buf(&client,port,buf)) exit(-1);
    if (leave_monitor(&client,port)) exit(-1);
  } else {
    printusage();
    exit(-1);
  }
  return 0;
}
